// (c) Thomas Kamps, 2009
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the7 implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.

#include <QLabel>
#include <QPainter>
#include <QStackedLayout>
#include <QMouseEvent>
#include <QAction>
#include <QMenu>
#include "field.h"
#include "ui_field.h"
#include "cmdtoolbutton.h"
#include "auction.h"
#include "src/Monopd/monopdfield.h"

Field::Field(QWidget* parent) : QWidget(parent), ui(new Ui::Field) {
	ui->setupUi(this);
	text = new QTextBrowser(this);
	auction = new Auction();
	stack = new QStackedLayout(ui->center);
	stack->setObjectName("StackField");
	stack->addWidget(text);
	stack->addWidget(auction);
	stack->setCurrentIndex(0);

	ui->wEstate->setVisible(false);
	ui->fButtons->setVisible(false);
	selectedField = 0;
	setMouseTracking(true);

	txtColor = false;
	

	acSell = new QAction(QIcon(":/images/tools/sell.png"), tr("Sell"), this);
	acSell->setData(QVariant(".es"));
	acMortage = new QAction(QIcon(":/images/tools/mortage.png"), tr("Mortgage"), this);
	acMortage->setData(QVariant(".em"));
	acUnmortage = new QAction(QIcon(":/images/tools/unmortage.png"), tr("Unmortgage"), this);
	acUnmortage->setData(QVariant(".em"));
	acBuyHouse = new QAction(QIcon(":/images/tools/buyhouse.png"), tr("Buy House"), this);
	acBuyHouse->setData(QVariant(".hb"));
	acSellHouse = new QAction(QIcon(":/images/tools/sellhouse.png"), tr("Sell House"), this);
	acSellHouse->setData(QVariant(".hs"));
	menu = new QMenu(this);

	//Under Win, StylePanel does not work
	#if defined(Q_WS_WIN)
	ui->centWidget->setFrameShape(QFrame::Panel);
	#endif
}

Field::~Field() {
	delete ui;
}

void Field::setGame(MonopdGame* g) {
	game = g;
}

void Field::setClientId(int id) {
	clientId = id;;
}

QRect Field::getAvatarRect(int pos) {
	QRect fRect = getFieldRect(pos);
	int side = 4*pos/N;

	int w;
	int h;

	if (side%2 == 0) {
		w = fRect.width()/2;
		h = fRect.height()/4;
	} else {
		w = fRect.width()/4;
		h = fRect.height()/2;
	}
	int s = w;
	if (h < w) s = h;

	return QRect(fRect.x()+(fRect.width()-s)/2, fRect.y()+(fRect.height()-s)/2, s, s);

	return QRect(0,0,0,0);
}

//calculates the rect for teh Field wiht id fid
QRect Field::getFieldRect(int fid) {

	int xOffset = (w-s)/2;
	int yOffset = (h-s)/2;

	int side = 4*fid/N;

	int fWidth = 0;
	int fHeight = 0;
	int fposX = 0;
	int fposY = 0;

	//get
	if (fid%(N/4) == 0) {
		//Edge
		fWidth = vertFieldWidth;
		fHeight = horFieldHeight;
	} else {
		if (side%2 == 0) {
			//We are on the North or south side
			fWidth = horFieldWidth;
			fHeight = horFieldHeight;
		} else {
			fWidth = vertFieldWidth;
			fHeight = vertFieldHeight;
		}
	}

	if (side == 0) {
		//South side
		fposX = vertFieldWidth+(N/4-1-fid)*horFieldWidth;
		fposY = cornerSize+(N/4-1)*vertFieldHeight;
	}
	if (side == 1) {
		//West side
		fposX = 0;
		fposY = horFieldHeight+(N/2-1-fid)*vertFieldHeight;
	}
	if (side == 2) {
		//North side
		if (fid > N/2) fposX = vertFieldWidth+(fid-N/2-1)*horFieldWidth;
		else fposX = 0;
		fposY = 0;
	}
	if (side == 3) {
		//East side
		fposX = cornerSize+(N/4-1)*horFieldWidth;
		if (fid > 3*N/4) fposY = horFieldHeight+(fid-3*N/4-1)*vertFieldHeight;
		else fposY = 0;
	}
	return QRect(fposX+xOffset, fposY+yOffset, fWidth, fHeight);
}

MonopdField* Field::getField(int x, int y) {
	for (int i = 0; i < game->getNumFields(); i++) {
		MonopdField* f = game->getField(i);
		if (f->isInRect(x, y)) return f;
	}
	return 0;
}

void Field::deleteMenu() {
	menu->clear();
}

void Field::changeEvent(QEvent *e) {
	QWidget::changeEvent(e);
	switch (e->type()) {
		case QEvent::LanguageChange:
			ui->retranslateUi(this);
			break;
		default:
			break;
	}
}

void Field::mousePressEvent(QMouseEvent* event) {
	if (game == 0) return;
	selectedField = getField(event->x(), event->y());
	if (event->button() == Qt::LeftButton) {
		deleteMenu();
		//Get the Field that got clicked
		if (selectedField == 0) emit(fieldSelected(0));
		else {
			if (selectedField->getGroup() == -1) {
				selectedField = 0;
				emit(fieldSelected(0));
			} else {
				emit(fieldSelected(selectedField));
			}
		}
		update();
	} else if (event->button() == Qt::RightButton) {
		if (selectedField == 0) return;
		if (selectedField->getGroup() == -1) return;
		if (selectedField->getOwner() != clientId) return;

		deleteMenu();
		if (game->getGroup(selectedField->getGroup())->getNumHouses() == 0) {
			menu->addAction(acSell);
			menu->addSeparator();
		}
		if (selectedField->getCanToggleMortage()) {
			if (selectedField->getMortaged()) menu->addAction(acUnmortage);
			else menu->addAction(acMortage);
			menu->addSeparator();
		}
		if (selectedField->getCanBuyHouse()) menu->addAction(acBuyHouse);
		if (selectedField->getCanSellHouse()) menu->addAction(acSellHouse);
		connect(menu, SIGNAL(triggered(QAction*)), this, SLOT(menuClick(QAction*)));
		menu->popup(QWidget::mapToGlobal(event->pos()));

	}
	event->accept();
}

void Field::mouseMoveEvent(QMouseEvent* event) {
	auction->setToolTip("");
	ui->centWidget->setToolTip("");
	if (game == 0) return;
	MonopdField* f = getField(event->x(), event->y());
	if (f == 0) {
		setToolTip("");
		return;
	}

	QString tooltip = "<span style=\"font-weight:bold\">"+f->getName()+"</span>";
	if (f->getBuyable() && (f->getOwner() == -1)) tooltip = tooltip+"<br>"+tr("Price")+": "+QString::number(f->getPrice());
	if (f->getOwner() != -1) {
		tooltip = tooltip+"<br>"+tr("Owner")+": "+game->getPlayer(f->getOwner())->getName();
		if (f->getCanToggleMortage() && !f->getMortaged()) tooltip = tooltip+"<br>"+tr("Mortgagevalue")+": "+QString::number(f->getMortageVal());
		if (f->getCanToggleMortage() && f->getMortaged()) tooltip = tooltip+"<br>"+tr("Unmortgagevalue")+": "+QString::number(f->getUnmortageVal());
	}
	if (f->getCanBuyHouse()) tooltip = tooltip+"<br>"+tr("Buy House")+": "+QString::number(f->getHousePrice());
	if (f->getCanSellHouse()) tooltip = tooltip+"<br>"+tr("Sell House")+": "+QString::number(f->getSellHousePrice());
	if (f->getMoney() > 0) tooltip = tooltip+"<br>"+tr("Fine")+": "+QString::number(f->getMoney());
	if (f->getColor().alpha() != 0) {
		QString txt[6];
		txt[0] = tr("Rent with no houses");
		txt[1] = tr("Rent with one house");
		txt[2] = tr("Rent with two houses");
		txt[3] = tr("Rent with three houses");
		txt[4] = tr("Rent with four houses");
		txt[5] = tr("Rent with a hotel");
		for (int i = 0; i < 6; i++) {
			QString preRent = "";
			QString postRent = "";
			if (f->getHouses() == i) {
				preRent = "<span style=\"font-weight:bold\">";
				postRent = "</span>";
			}
			int acRent = f->getRent(i);

			//If a group is ownes by one single player, the normale rent is doubled
			if ((i == 0) && game->getGroup(f->getGroup())->getAllEstatesSameOwner()) acRent *= 2;
			tooltip = tooltip+"<br>"+txt[i]+": "+preRent+QString::number(acRent)+postRent;
		}
	}
	setToolTip(tooltip);
}

void Field::emitButtonClick(QString cmd) {
	emit(buttonClicked(cmd));
	clearButtons();
}

void Field::menuClick(QAction* a) {
	if (selectedField != 0) {
		emit(buttonClicked(a->data().toString()+QString::number(selectedField->getId())));
	}
	emit(fieldSelected(0));
	selectedField = 0;
	deleteMenu();
}

void Field::showField(int fid) {
	if (game == 0) return;
	QString name = "";
	QString price = "";
	if (fid == -1) {
		//clear Estate
		ui->wEstate->setVisible(false);
	} else {
		MonopdField* f = game->getField(fid);
		name = f->getName();
		price = QString::number(f->getPrice());
		int r = f->getColor().red();
		int g = f->getColor().green();
		int b = f->getColor().blue();
		int v = f->getColor().value();
		QString fColor = "#000000";
		if (v < 128) fColor = "#ffffff";
		QString sr = QString::number(r, 16);
		QString sg = QString::number(g, 16);
		QString sb = QString::number(b, 16);
		if (sr.size() < 2) sr = "0"+sr;
		if (sg.size() < 2) sg = "0"+sg;
		if (sb.size() < 2) sb = "0"+sb;
		QString col = "#"+sr+sg+sb;
		if (f->getGroup() != -1) {
			if (f->getColor().alpha() == 0) {
				ui->lEstate->setText("<span style=\"font-size:20pt; font-weight:bold\">"+name+"</span>");
			} else {
				ui->lEstate->setText("<span style=\"background:"+col+"; font-size:20pt; font-weight:bold; color:"+fColor+"\">"+name+"</span>");
			}
			ui->lPrice->setText("<span style=\"font-size:20pt; font-weight:bold\">"+price+"</span>");
		} else {
			ui->lEstate->setText("<span style=\"font-size:20pt; font-weight:bold\">"+name+"</span>");
			ui->lPrice->setText("");
		}
		ui->lEstate->setToolTip(name);
		ui->lPrice->setToolTip(price);
		ui->wEstate->setVisible(true);
	}
}

void Field::appendLine(QString txt) {
	QString bg = "#ffffff";
	QString color = "#000000";

	if (txtColor) bg = "#f0f0f0";
	txtColor =!txtColor;

	txtHtml = txtHtml+"<p style=\"margin-top:0px; margin-bottom:0px; margin-left:0px; margin-right:0px; -qt-block-indent:0; text-indent:0px; background:"+bg+"\"><span style=\"color:"+color+"\">"+txt+"</span></p>";
	text->setHtml(txtHtml+"<a name=\"end\" style=\"font-size:1px\">&nbsp;</a>");
	text->scrollToAnchor("end");

	text->setToolTip("");
}

void Field::addButton(QString text, QString cmd) {
	CmdToolButton* b = new CmdToolButton();
	b->setText(text);
	b->cmd = cmd;
	if (cmd == ".eb") b->setIcon(QIcon(":/images/field/document.png"));
	if (cmd == ".ea") b->setIcon(QIcon(":/images/tools/auction.png"));
	if (cmd == ".E") b->setIcon(QIcon(":/images/tools/endturn.png"));

	if (cmd == ".jr") b->setIcon(QIcon(":/images/tools/jailroll.png"));
	if (cmd == ".jc") b->setIcon(QIcon(":/images/tools/jailcard.png"));
	if (cmd == ".jp") b->setIcon(QIcon(":/images/tools/jailpay.png"));

	if (cmd == ".T$") b->setIcon(QIcon(":/images/tools/jailpay.png"));
	if (cmd == ".T%") b->setIcon(QIcon(":/images/tools/paypercent.png"));

	b->setToolTip(b->text());
	b->setIconSize(QSize(24,24));
	connect(b, SIGNAL(cmdClicked(QString)), this, SLOT(emitButtonClick(QString)));
	if (cmd == ".gx") clearButtons();
	buttons.append(b);
	ui->buttons->insertWidget(buttons.size(), b);
	ui->fButtons->setVisible(true);
}

void Field::clearText() {
	text->setHtml("");
	txtColor = false;
	txtHtml = "";
	text->setToolTip("");
	clearButtons();
}

void Field::clearButtons() {
	while (buttons.size() > 0) {
		CmdToolButton* b = buttons.value(0);
		buttons.removeAll(b);
		delete b;
	}
	ui->fButtons->setVisible(false);
}

void Field::addPlayerAuction(Player* p) {
	auction->addPlayerWidget(p->auction);
}

void Field::showAuction() {
	if (game == 0) return;
	stack->setCurrentIndex(1);
	for (int i = 0; i < game->getNumPlayer(); i++) {
		Player* pl = game->getPlayerFromList(i);
		pl->auction->setName(pl->getName());
		pl->auction->setMoney("0");
	}
	auction->resetSpinBox();
}

void Field::updatAuctionId(int aid) {
	auction->setId(aid);
}

void Field::updateAuctionField(int fid) {
	if (game == 0) return;
	auction->setFieldName(game->getField(fid)->getName());
}

void Field::bidUpdate(int pid, int bid) {
	if (game == 0) return;
	Player* p = game->getPlayer(pid);
	if (p == 0) return;
	for (int i = 0; i < game->getNumPlayer(); i++) {
		Player* pl = game->getPlayerFromList(i);
		pl->auction->setName(pl->getName());
	}
	p->auction->setName("<b>"+p->getName()+"</b>");
	p->auction->setMoney(QString::number(bid));
	auction->setMinBid(bid+1);
}

void Field::updateAuctionStatus(int status) {
	switch (status) {
		case 0:
			auction->setStatusText("");
			break;
		case 1:
			auction->setStatusText(tr("First..."));
			break;
		case 2:
			auction->setStatusText(tr("Second..."));
			break;
		case 3:
			stack->setCurrentIndex(0);
			break;
		default:
			break;
	}
}

void Field::paintEvent(QPaintEvent* pe) {
	if (game == 0) return;
	N = game->getNumFields();
	w = this->width()-1;
	h = this->height()-1;
	s = w;
	if (h < w) s = h;

	QPainter p(this);

	//Set size of widget in Field-center
	cornerSize = 2*s/13;
	vertFieldHeight = (s-2*cornerSize)/((N/4)-1);
	horFieldWidth = vertFieldHeight;
	horFieldHeight = cornerSize;
	vertFieldWidth = cornerSize;

	s = ((N/4)-1)*horFieldWidth+2*cornerSize;

	//Now draw South-Fields, we draw from left to right
	for (int i = 0; i < N; i++) {
		int side = 4*i/N;

		//Get Data for drawing
		MonopdField* f = game->getField(i);
		QRect fRect = getFieldRect(i);
		f->setFieldRect(fRect);

		int fposX = fRect.x();
		int fposY = fRect.y();
		int fWidth = fRect.width();
		int fHeight = fRect.height();

		QRect strRect; //Drawing rect for the street-color
		QRect fName; //Rect for the Fieldname

		QRect docRect; //Rect for document or owner-avatar
		int houseStartX = 0; //  Position of the houses
		int houseStartY = 0; //
		int houseWidth = 0;  //   Sizes of a house
		int houseHeight = 0; //
		int houseOffsetX = 0; //  Offset for house drawing
		int houseOffsetY = 0; //
		if (side == 0) {
			//South side
			int docS =  min(fWidth,fHeight/8);
			strRect = QRect(fposX, fposY, fWidth, fHeight/4);
			fName = QRect(fposX, fposY+fHeight/4, fWidth, 5*fHeight/8);
			docRect = QRect(fposX+fWidth-docS, fposY+fHeight-docS, docS, docS);
			houseStartX = fposX,
			houseStartY = fposY+2;
			houseWidth = fWidth/4;
			houseHeight = fHeight/4-4;
			houseOffsetX = fWidth/4;
			houseOffsetY = 0;
		} else if (side == 1) {
			//West side
			int docS =  min(fWidth/8,fHeight);
			strRect = QRect(fposX+3*fWidth/4, fposY, fWidth/4, fHeight);
			fName = QRect(fposX, fposY, 3*fWidth/4, 3*fHeight/4);
			docRect = QRect(fposX, fposY+fHeight-docS, docS, docS);
			houseStartX = fposX+3*fWidth/4+2,
			houseStartY = fposY;
			houseWidth = fWidth/4-4;
			houseHeight = fHeight/4;
			houseOffsetX = 0;
			houseOffsetY = fHeight/4;
		} else if (side == 2) {
			//North side
			int docS =  min(fWidth,fHeight/8);
			strRect = QRect(fposX, fposY+3*fHeight/4, fWidth, fHeight/4);
			fName = QRect(fposX, fposY+fHeight/8, fWidth, 5*fHeight/8);
			docRect = QRect(fposX, fposY, docS, docS);
			houseStartX = fposX,
			houseStartY = fposY+3*fHeight/4+2;
			houseWidth = fWidth/4;
			houseHeight = fHeight/4-4;
			houseOffsetX = fWidth/4;
			houseOffsetY = 0;
		} else if (side == 3) {
			//East side
			int docS =  min(fWidth/8,fHeight);
			strRect = QRect(fposX, fposY, fWidth/4, fHeight);
			fName = QRect(fposX+fWidth/4, fposY, 5*fWidth/8, fHeight);
			docRect = QRect(fposX+fWidth-docS, fposY, docS, docS);
			houseStartX = fposX+2,
			houseStartY = fposY;
			houseWidth = fWidth/4-4;
			houseHeight = fHeight/4;
			houseOffsetX = 0;
			houseOffsetY = fHeight/4;
		}

		if (f != 0) {
			if (f->getOwner() != -1) {
				//Draw light green, if field is owned
				p.setBrush(QColor(192,255,192));
			} else {
				//Otherwise make white
				p.setBrush(QColor(255,255,255));
			}
			p.drawRect(fposX, fposY, fWidth, fHeight);

			if (f->getBuyable()) {
				//Draw document or avatar of owner
				QPixmap pix;
				if (f->getOwner() == -1) {
					pix = QPixmap(":/images/field/document.png");
				} else {
					pix = QPixmap(":/images/avatars/"+game->getPlayer(f->getOwner())->getAvatar());
				}
				p.drawPixmap(docRect, pix);
				if (f->getMortaged()) {
					p.setBrush(QColor(0,0,0,128));
					p.drawRect(fposX, fposY, fWidth, fHeight);
				}
			}

			//Draw field name
			p.setPen(Qt::black);
			p.drawText(fName, Qt::AlignCenter|Qt::TextWrapAnywhere, f->getName());

			if (f->getGroup() != -1) {
				//Draw streetColor
				p.setPen(f->getColor());
				p.setBrush(f->getColor());
				p.drawRect(strRect);

				//draw Houses or Hotel
				p.setPen(Qt::black);
				if (f->getHouses() < 5) {
					p.setBrush(Qt::green);
					for (int k = 0; k < f->getHouses(); k++) {
						p.drawRect(houseStartX+k*houseOffsetX, houseStartY+k*houseOffsetY, houseWidth, houseHeight);
					}
				} else {
					p.setBrush(Qt::red);
					p.drawRect(houseStartX, houseStartY, houseWidth+houseOffsetX, houseHeight+houseOffsetY);
				}
			}
		}

		//Draw Field Borders
		p.setPen(QColor(0,0,0));
		p.setBrush(Qt::NoBrush);
		p.drawRect(fposX, fposY, fWidth, fHeight);
	}

	//Draw selectedField
	if (selectedField != 0) {
		p.setBrush(QColor(128,128,255,64));
		p.drawRect(selectedField->getFieldRect());
	}

	//Draw Players
	for (int i = 0; i < game->getNumPlayer(); i++) {
		Player* pl = game->getPlayerFromList(i);
		if (pl == 0) continue;
		if (pl->getGame() == game->getGameId()) {
			QRect avRect = getAvatarRect(pl->getPos());
			QPixmap pix = QPixmap(":/images/avatars/"+pl->getAvatar());
			p.drawPixmap(avRect, pix);
			if (pl->getInJail()) p.drawPixmap(avRect, QPixmap(":/images/field/jailed.png"));
		}
	}

	p.end();

	ui->centWidget->setMaximumWidth((N/4-1)*horFieldWidth-20);
	ui->centWidget->setMaximumHeight((N/4-1)*vertFieldHeight-20);
	ui->centWidget->setMinimumWidth((N/4-1)*horFieldWidth-20);
	ui->centWidget->setMinimumHeight((N/4-1)*vertFieldHeight-20);

	pe->accept();
}
